__version__ = '0.1.0'
from . lark_sql import MongoSQLParser
from . mongodb_connector import *

__all__ = ['connect','MongoDBConnectorException', 'MongoSQLParser', 
'Warning', 'Error', 'InterfaceError','DataError','DatabaseError','OperationalError','IntegrityError','InternalError','ProgrammingError','NotSupportedError','MongoDBConnectorException']

"""

InterfaceError
   |__DatabaseError
      |__DataError
      |__OperationalError
      |__IntegrityError
      |__InternalError
      |__ProgrammingError
      |__NotSupportedError

"""


import re
from pymongo import MongoClient
from . lark_sql import MongoSQLParser

class MongoDBConnectorException(Exception):
    pass

class Warning(Exception):
    pass

class Error(Exception):
    pass

class InternalError(Error):
    pass

class DatabaseError(Error):
    pass

class DataError(DatabaseError):
    pass

class OperationError(DatabaseError):
    pass

class IntegrityError(DatabaseError):
    pass

class InternalError(DatabaseError):
    pass

class ProgrammingError(DatabaseError):
    pass

class NotSupportedError(DatabaseError):
    pass

"""

InterfaceError
Exception raised for errors that are related to the database interface rather than the database itself. It must be a subclass of Error.
DatabaseError
Exception raised for errors that are related to the database. It must be a subclass of Error.
DataError
Exception raised for errors that are due to problems with the processed data like division by zero, numeric value out of range, etc. It must be a subclass of DatabaseError.
OperationalError
Exception raised for errors that are related to the database's operation and not necessarily under the control of the programmer, e.g. an unexpected disconnect occurs, the data source name is not found, a transaction could not be processed, a memory allocation error occurred during processing, etc. It must be a subclass of DatabaseError.
IntegrityError
Exception raised when the relational integrity of the database is affected, e.g. a foreign key check fails. It must be a subclass of DatabaseError.
InternalError
Exception raised when the database encounters an internal error, e.g. the cursor is not valid anymore, the transaction is out of sync, etc. It must be a subclass of DatabaseError.
ProgrammingError
Exception raised for programming errors, e.g. table not found or already exists, syntax error in the SQL statement, wrong number of parameters specified, etc. It must be a subclass of DatabaseError.
NotSupportedError
Exception raised in case a method or database API was used which is not supported by the database, e.g. requesting a .rollback() on a connection that does not support transaction or has transactions turned off. It must be a subclass of DatabaseError.
This is the exception inheritance layout:

StandardError
|__Warning
|__Error
   |__InterfaceError
   |__DatabaseError
      |__DataError
      |__OperationalError
      |__IntegrityError
      |__InternalError
      |__ProgrammingError
      |__NotSupportedError
"""


class MongoDBCursor():

    arraysize = 100
    """This read/write attribute specifies the number of rows to fetch at a time with .fetchmany(). It defaults to 1 meaning to fetch a single row at a time.
    Implementations must observe this value with respect to the .fetchmany() method, but are free to interact with the database a single row at a time. It may also be used in the implementation of .executemany().
    """

    def __init__(self, client , database):
        self.client = client
        self.database = database
        self.parser = MongoSQLParser.get_instance()
        print("MongoDB Cursor created")

    def callproc(procname, *args):
        raise NotSupportedError("Not implemented")

    def close(self):
        self.client.close()

    def execute(self, operation, *args):
        print("execute", operation)

        try:
            parsed = self.parser.parse(operation)
            if parsed["statement"] == "select":
                print("select")
                collection = parsed["table"]
                print(collection)
                print(parsed["where"])
                self.cursor = self.client[self.database][collection].find(parsed["where"])
                print("select done.")
            elif parsed["statement"] == "insert":
                print("insert")
                collection = parsed["table"]
                k = parsed["columns"] # name, year
                v = parsed["values"]  # mike 30
                obj = {}
                for i in range(len(k)):
                    obj[ k[i] ] = v[i]
                self.client[self.database][collection].insert_one(obj)
                self.cursor = None
                print("insert done.")
            elif parsed["statement"] == "update":
                print("update")
                collection = parsed["table"]
                self.client[self.database][collection].update_many(parsed["where"],{'$set': parsed["set"]})
                self.cursor =  None
                print("update done!!")
            elif parsed["statement"] == "delete":
                print("delete")
                collection = parsed["table"]
                print("parsed[where]", parsed["where"])
                # self.client[self.database][collection].delete_many(parsed["where"])
                # {"name": {"$eq": "Charly"}}
                self.client[self.database][collection].delete_many({"name": {"$eq": "Charly"}})
                self.cursor = None
            elif parsed["statement"] == "create":
                collection = parsed["table"]
                self.client[self.database][collection]
            elif parsed["statement"] == "drop":
                collection = parsed["table"]
                self.client[self.database][collection].drop()
            else:
                raise NotSupportedError("Not implemented")
        except Exception as e:
            raise OperationError("Invalid SQL statement", e)

    def executemany( self, operation, seq_of_parameter ):
        pass

    def fetchone(self):
        if self.cursor is not None:
            return self.cursor.next()
        else:
            return None

    def fetchmany(self):
        """ Fetch the next set of rows of a query result, returning a sequence of sequences (e.g. a list of tuples). An empty sequence is returned when no more rows are available.
            The number of rows to fetch per call is specified by the parameter. If it is not given, the cursor's arraysize determines the number of rows to be fetched. The method should try to fetch as many rows as indicated by the size parameter. If this is not possible due to the specified number of rows not being available, fewer rows may be returned.
            An Error (or subclass) exception is raised if the previous call to .execute*() did not produce any result set or no call was issued yet.
            Note there are performance considerations involved with the size parameter. For optimal performance, it is usually best to use the .arraysize attribute. If the size parameter is used, then it is best for it to retain the same value from one .fetchmany() call to the next.
        """
        ret = []
        if self.cursor is not None:
            for i in range(self.arraysize):
                try:
                    ret.append(self.cursor.next())
                except StopIteration:
                    break
        return ret

    def fetchall(self):
        """Fetches all (remaining) rows of a query result, returning them as a sequence of sequences (e.g. a list of tuples). Note that the cursor's arraysize attribute can affect the performance of this operation.
            An Error (or subclass) exception is raised if the previous call to .execute*() did not produce any result set or no call was issued yet.
        """
        ret = []
        if self.cursor is not None:
            for i in self.cursor:
                ret.append(i)
        return ret

    def setinputsizes(self, sizes):
        """
            This can be used before a call to .execute*() to predefine memory areas for the operation's parameters.
            sizes is specified as a sequence — one item for each input parameter. The item should be a Type Object that corresponds to the input that will be used, or it should be an integer specifying the maximum length of a string parameter. If the item is None, then no predefined memory area will be reserved for that column (this is useful to avoid predefined areas for large inputs).
            This method would be used before the .execute*() method is invoked.
            Implementations are free to have this method do nothing and users are free to not use it.
        """
        pass


    def setoutputsize(self, size, column):
        """
            Set a column buffer size for fetches of large columns (e.g. LONGs, BLOBs, etc.). The column is specified as an index into the result sequence. Not specifying the column will set the default size for all large columns in the cursor.
            This method would be used before the .execute*() method is invoked.
            Implementations are free to have this method do nothing and users are free to not use it.
        """
        pass



class MongoDBConnection():
    def __init__(self,host:str,port:int=None, user:str=None, password:str=None, database:str=None):
        self.host = host

        if port is not None:
            try:
                self.port = int(port)
            except Exception as e:
                raise ProgrammingError("Invalid port number", e)
        else:
            self.port = 27017

        self.user = user
        self.password = password
        self.database = database

        if self.database is None:
            raise ProgrammingError("Database name is not specified")
        
        user_pass =""
        if user is not None:
            user_pass = user + ":" + password + "@"

        uri = "mongodb://" + user_pass + host + ":" + str(port)
        try:
            self.client = MongoClient(uri)
        except Exception as e:
            raise ProgrammingError("Invalid mongodb connection configuration", e)
    
    def close(self):
        self.client.close()
        print("MongoDB Connection closed")

    def commit(self):
        pass

    def rollback(self):
        pass

    def cursor(self)->MongoDBCursor:
        return MongoDBCursor(self.client, self.database)



def connect(uri:str)->MongoDBConnection:
    """Connect to MongoDB
    :param uri: MongoDB URI (e.g. mongodb://user:password@localhost:27017/db)
    :return: MongoDBConnection
    """
    match = re.match(r'^mongodb://((?P<user>.*):(?P<password>.*)@)?(?P<host>.*?)(:(?P<port>[1-9][0-9]*))?/(?P<database>.*)$', uri)
    d = match.groupdict()
    return MongoDBConnection(d["host"], int(d["port"]), d["user"], d["password"], d["database"])



apilevel = 2.0
threadsafety = 0
paramstyle = "qmark"


"""
Connection methods
.close()
Close the connection now (rather than whenever .__del__() is called).

The connection will be unusable from this point forward; an Error (or subclass) exception will be raised if any operation is attempted with the connection. The same applies to all cursor objects trying to use the connection. Note that closing a connection without committing the changes first will cause an implicit rollback to be performed.

.commit()
Commit any pending transaction to the database.

Note that if the database supports an auto-commit feature, this must be initially off. An interface method may be provided to turn it back on.

Database modules that do not support transactions should implement this method with void functionality.

.rollback()
This method is optional since not all databases provide transaction support. [3]

In case a database does provide transactions this method causes the database to roll back to the start of any pending transaction. Closing a connection without committing the changes first will cause an implicit rollback to be performed.

.cursor()
Return a new Cursor Object using the connection.

If the database does not provide a direct cursor concept, the module will have to emulate cursors using other means to the extent needed by this specification. [4]

"""

"""
Constructors
Access to the database is made available through connection objects. The module must provide the following constructor for these:

connect( parameters... )
Constructor for creating a connection to the database.

Returns a Connection Object. It takes a number of parameters which are database dependent. [1]

Globals
These module globals must be defined:

apilevel
String constant stating the supported DB API level.

Currently only the strings "1.0" and "2.0" are allowed. If not given, a DB-API 1.0 level interface should be assumed.

threadsafety
Integer constant stating the level of thread safety the interface supports. Possible values are:

threadsafety	Meaning
0	Threads may not share the module.
1	Threads may share the module, but not connections.
2	Threads may share the module and connections.
3	Threads may share the module, connections and cursors.
Sharing in the above context means that two threads may use a resource without wrapping it using a mutex semaphore to implement resource locking. Note that you cannot always make external resources thread safe by managing access using a mutex: the resource may rely on global variables or other external sources that are beyond your control.

paramstyle
String constant stating the type of parameter marker formatting expected by the interface. Possible values are [2]:

paramstyle	Meaning
qmark	Question mark style, e.g. ...WHERE name=?
numeric	Numeric, positional style, e.g. ...WHERE name=:1
named	Named style, e.g. ...WHERE name=:name
format	ANSI C printf format codes, e.g. ...WHERE name=%s
pyformat	Python extended format codes, e.g. ...WHERE name=%(name)s
"""


import logging
from lark import Lark, logger, Transformer
from lark.lexer import Token
import mongodb_connector

"""
?start: value
 
?value: object
        | array
        | string
        | SIGNED_NUMBER      -> number
        | "true"             -> true
        | "false"            -> false
        | "null"             -> null
 
array  : "[" [value ("," value)*] "]"
object : "{" [pair ("," pair)*] "}"
pair   : string ":" value
 
string : ESCAPED_STRING
 
%import common.ESCAPED_STRING
%import common.SIGNED_NUMBER
%import common.WS
 
%ignore WS
"""

GRAMMAR = '''

?start: select_statement|insert_statement|update_statement|delete_statement|create_statement|drop_statement

//ルール
select_statement: SELECT columns [FROM table_name] [where_clause] [order_by_clause] [limit_clause]

//INSERT INTO テーブル名 (列名1, 列名2,...) VALUES (値1, 値2,...);
insert_statement: INSERT INTO table_name "(" columns ")" VALUES "(" values ")"

// UPDATE (表名) SET (カラム名1) = (値1), (カラム名2) = (値2) WHERE (条件);
update_statement: UPDATE table_name SET assigns [where_clause]

// DELETE (表名) FROM (テーブル名) WHERE (条件);
delete_statement: DELETE FROM table_name [where_clause]

//CREATE TABLE (テーブル名) (列名1 型1, 列名2 型2,...)
create_statement: CREATE TABLE table_name "(" COLUMN type ("," COLUMN type)* ")"

//DROP TABLE (テーブル名)
drop_statement: DROP TABLE table_name

columns:  ANY | COLUMN ("," COLUMN)*
values: value ("," value)*
order_by_clause: ORDER BY COLUMN [ASC|DESC]
where_clause: WHERE expression
limit_clause: LIMIT INT

expression: and_condition [OR and_condition]
and_condition: condition (AND condition)*
condition: compare | "(" expression ")"
compare: term COMPARE term

assigns: (COLUMN "=" value) ("," COLUMN "=" value)*

value: SIGNED_INT | SIGNED_FLOAT | STRING  |NULL | boolean
term: value | COLUMN
boolean: TRUE | FALSE
TRUE.2: "true"|"True"|"TRUE"
FALSE.3: "false"|"False"|"FALSE"
NULL.5: "NULL"|"null"|"Null"


//終端記号
table_name: WORD
ANY.10: "*"
COLUMN.1: WORD
STRING: ESCAPED_STRING

SELECT: "SELECT"|"select"
FROM: "FROM"|"from"
WHERE: "WHERE"|"where"
ORDER: "ORDER"|"order"
BY: "BY"|"by"
ASC: "ASC"|"asc"
DESC: "DESC"|"desc"
LIMIT: "LIMIT"|"limit"
INSERT: "INSERT"|"insert"
INTO: "INTO"|"into"
VALUES: "VALUES"|"values"
UPDATE: "UPDATE"|"update"
SET: "SET"|"set"
DELETE: "DELETE"|"delete"
OR: "OR"|"or"
AND: "AND"|"and"
NOT: "NOT"|"not"
COMPARE:"<>"|"<="|">="|"<"|">"|"="|"!="
CREATE: "CREATE"|"create"
DROP: "DROP"|"drop"
TABLE: "TABLE"|"table"

type:  TYPE_BOOLEAN| TYPE_INT | TYPE_DECIMAL | TYPE_VARCHAR | TYPE_TIME

// 型情報　ただし使われません
TYPE_BOOLEAN: "BOOLEAN"|"boolean"
TYPE_INT: "INT"|"int"
TYPE_DECIMAL: "DECIMAL"|"decimal"
TYPE_VARCHAR: "VARCHAR"|"varchar"
TYPE_TIME: "TIME"|"time"

// サポートできていないもの
// IS: "IS"|"is"
// LIKE: "LIKE"|"like"
// IN: "IN"|"in"
// BETWEEN: "BETWEEN"|"between"
// JOIN: "JOIN"|"join"

//共通記号
%import common.ESCAPED_STRING
%import common.SIGNED_FLOAT
%import common.SIGNED_INT
%import common.INT
%import common.WORD
%import common.WS
%ignore WS
'''

#  FROM_CLAUSE WHERE_CLAUSE
# %import common.SIGNED_NUMBER

# pymongo cursor


class SQLTransformer(Transformer):

    def __clear(self):
        self.parsed_data = {}

    def __init__(self):
        self._functions = {}
        self.parsed_data = {}


    def select_statement(self, parsed):
        print("Select statement", parsed)
        ret = {
            "statement": "select",
            "table": parsed[3],
            "columns": parsed[1],
        }
        # SELECT[0] *[1] FROM[2] table[3] [WHERE[4] condition[5]] [ORDER BY[6] column[7] [ASC|DESC[8]]] [LIMIT[9] limit[10]]

        for i in range(4, len( parsed )):
            clause = parsed[i][0]
            ret[clause] = parsed[i][1]
        print(ret)
        return ret

    def columns(self, parsed):
        print("columns ", parsed)
        if type(parsed[0]) == Token and parsed[0].type == "ANY":
            return [parsed[0].value]
        else:
            ret = []
            for i in range(len(parsed)):
                ret.append(parsed[i].value)
            return ret

    # exexpression: and_condition [OR and_condition]
    def expression(self, parsed):
        if len( parsed ) == 3:
            # OR condition
            return {
                "$or":[
                    parsed[0],
                    parsed[2]
                ]
            }
        else:
            return parsed[0]

    # condition: compare | "(" expression ")"
    def condition(self, parsed):
        if len(parsed) == 1:
            return parsed[0]
        else:
            return parsed[1] # parsed expression

    # and_condition: condition (AND condition)*
    def and_condition(self, parsed):
        print("And condition", parsed)
        if len( parsed ) >= 3:
            # AND condition
            return {
                [parsed[i] for i in range(len(parsed)) if i % 2 == 0]
            }
        else:
            print("return parsed[0]",parsed[0])
            return parsed[0]

    def where_clause(self,paserd):
        print("Where" , paserd)
        paserd[0] = "where"
        return paserd

    def __get_column(self, token):
        if type(token) is Token and token.type == "COLUMN":
            print("__get_column", token.value)
            return token.value
        return None
    

    # COMPARE:"<>"|"<="|">="|"<"|">"|"="|"!="
    def __to_mongo_condition(self, obj ):
        print(obj,)
        if obj["op"] == "<>" or obj["op"] == "!=":
            return {
                obj["left"]["value"]  : { "$ne": obj["right"]["value"] }
            }
        elif obj["op"] == "<=":
            return {
                obj["left"]["value"]  : { "$lte": obj["right"]["value"] }
            }
        elif obj["op"] == ">=":
            return {
                obj["left"]["value"] : { "$gte": obj["right"]["value"] }
            }
        elif obj["op"] == "<":
            return {
                obj["left"]["value"]: { "$lt": obj["right"]["value"] }
            }
        elif obj["op"] == ">":
            return {
                obj["left"]["value"]: { "$gt": obj["right"]["value"] }
            }
        elif obj["op"] == "=":
            return {
                obj["left"]["value"]: { "$eq": obj["right"]["value"] }
            }
        else:
            raise ValueError("Unknown operator")
    

    def compare(self, tokens):
        print("compare ", tokens, type(tokens[0]), type(tokens[1]), type(tokens[2]))

        ret = {
            "left": {},
            "op": None,
            "right": {}
        }

        c = self.__get_column(tokens[0])
        if c is None:
            ret["left"]["value"] = tokens[0]
        else:
            ret["left"] = {
                "type": "column",
                "value": c
            }

        # "<>"|"<="|">="|"<"|">"|"="|"!="
        ret["op"] = tokens[1].value
        
        c = self.__get_column(tokens[2])
        if c is None:
            ret["right"]["value"] = tokens[2]
        else:
            ret["right"] = {
                "type": "column",
                "value": c
            }

        v = self.__to_mongo_condition(ret)
        return v
    
    def order_by_clause(self,parsed):
        print("order by ",parsed)
        ret = {
            "sort": parsed[2].value,
        }
        if len(parsed)>2:
            ret["sort_order"] = parsed[3].value
        return ["order_by", ret]

    def limit_clause(self,tokens):
        print("limit", tokens)
        return ["limit", tokens[1].value]

    def insert_statement(self, parsed):
        print("Insert statement", parsed)
        ret ={
            "statement": "insert",
            "table": parsed[2],
            "columns": parsed[3],
            "values": parsed[5]
        }
        print(ret)
        return ret
    
    def values(self, parsed):
        print("Values", parsed)
        return parsed
    
    def update_statement(self, parsed):
        # UPDATE table SET assigns [where_clause]
        print("Update statement", parsed )

        ret = {
            "statement": "update",
            "table": parsed[1],
            "set": parsed[3]
        }

        if len(parsed) == 5:
            ret["where"] = parsed[4][1]

        return ret
    
    def assigns(self,parsed):
        ret = {}
        for i in range(0, len(parsed), 2):
            print(parsed[i], parsed[i+1])
            ret[parsed[i].value] = parsed[i+1]
        return ret

    def delete_statement(self, parsed):
        print("Delete statement", parsed)
        ret = {
            "statement": "delete",
            "table": parsed[2],
        }
        if len(parsed) > 3:
            ret["where"] = parsed[3][1]
        print(ret)
        return ret

    def create_statement(self, parsed):
        print("Create statement",parsed)
        ret = {
            "statement": "create",
            "table": parsed[2]
        }
        return ret


    def drop_statement(self, parsed):
        print("Drop statement")
        ret = {
            "statement": "drop",
            "table": parsed[2]
        }
        return ret

    def table_name(self, tree):
        print("table_func " , type(tree[0].value), tree[0])
        return tree[0].value

    def boolean(self, tokens):
        b = tokens[0].value
        ret = bool(b.upper())
        return ret

    def null(self, token ):
        return None
    
    def value(self, t):
        if type(t) is list:
            if t[0].type=='SIGNED_INT':
                return int(t[0].value)
            elif t[0].type == 'SIGNED_FLOAT':
                return float(t[0].value)
            elif t[0].type == 'STRING':
                s = t[0].value[1:-1]
                print("STRING", t[0].value, s)
                return s
            else:
                return t[0].value
        else:
            return t
    
    def term(self, t):
        return t[0]

class MongoSQLParser():

    def __init__(self):
        self.parser = Lark(GRAMMAR, parser='lalr', debug=True, transformer=SQLTransformer())

    @classmethod
    def get_instance(cls)->'MongoSQLParser':
        if not hasattr(cls, "_instance"):
            cls._instance = MongoSQLParser()
        return cls._instance
    
    def parse(self, statement:str):
        try:
            return self.parser.parse(statement)
        except Exception as e:
            raise mongodb_connector.MongoDBConnectorException("Parse Failed.", statement)

if __name__ == "__main__":
    print("lark sql") 
    print(GRAMMAR)

    # logger.setLevel(logging.DEBUG)
    # p = Lark(GRAMMAR, parser='lalr', debug=True, transformer=SQLTransformer())
    parser = MongoSQLParser.get_instance()

    # select
    # print(p.parse('SELECT hoge,hage FROM hoge WHERE ((hoge > 10)AND(hoge <10)OR(c>10))'))
    # print(p.parse('SELECT hoge,hage FROM hoge WHERE hoge > 10 AND hoge <10 OR x > -10'))
    # print(p.parse('SELECT hoge,hage FROM hoge WHERE hoge  != 20 ORDER BY hoge ASC LIMIT 10'))
    # print(p.parse('SELECT * FROM csv WHERE year!= 20 ORDER BY year ASC LIMIT 10'))
    # parsed = parser.parse('SELECT * FROM csv WHERE year!= 20 ORDER BY year ASC LIMIT 10')


    # # inseret
    # print(p.parse('INSERT INTO hoge (a, b, c) VALUES (1,2,3)'))
    # parsed = parser.parse('INSERT INTO hoge (a, b, c) VALUES (1,2,3)')
 
    # # delete
    # print(p.parse('DELETE FROM hoge WHERE x > 10'))
    # print(parser.parse('DELETE FROM hoge WHERE x > 10'))

    # # update
    print(parser.parse('UPDATE hoge SET a=19, b=20, c=30 WHERE x > 20'))

    # create table
    # print(p.parse('CREATE TABLE hoge (a INT, b INT, c INT)'))

    # drop table
    # print(p.parse('DROP TABLE hoge'))
    # https://docs.mongodb.com/manual/reference/method/db.collection.drop/


